/*

  Copyright (c) 2016 Steffen Illhardt

  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
  the following conditions:

  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#define _CRT_SECURE_NO_WARNINGS
#define WIN32_LEAN_AND_MEAN

#ifdef _WIN32_WINNT
# undef _WIN32_WINNT
#endif
#define _WIN32_WINNT 0x600

#ifndef UNICODE
# define UNICODE
#endif

#if defined(UNICODE) && !defined(_UNICODE)
# define _UNICODE
#elif defined(_UNICODE) && !defined(UNICODE)
# undef _UNICODE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <tchar.h>
#include <errno.h>
#include <windows.h>
#include <tlhelp32.h>

typedef struct tag_CALLBACKDATA
{
  HWND hwnd;
  DWORD pid;
  DWORD tid;
  TCHAR *title;
  TCHAR name[MAX_PATH];
} CALLBACKDATA;

BOOL PnameToPid(CALLBACKDATA *const pData);
BOOL PidToPname(CALLBACKDATA *const pData);
BOOL CALLBACK EnumWindowsListCallback(HWND hwnd, LPARAM lParam);
BOOL CALLBACK EnumWindowsPIDCallback(HWND hwnd, LPARAM lParam);
BOOL CALLBACK EnumWindowsTIDCallback(HWND hwnd, LPARAM lParam);
BOOL CALLBACK EnumWindowsTitleCallback(HWND hwnd, LPARAM lParam);
void PrintData(const CALLBACKDATA *const pData);
void Help(TCHAR *const owncall);

int main(void)
{
  /*** working around missing _tmain/wmain support of MinGW (shall also work with other compilers) ***/
  int argc = 0;
  TCHAR **argv = NULL, **envp = NULL;
  {
    struct _startupinfo { int newmode; } si = {0};
    static const char crtFileName[] = "msvcrt";
    HMODULE crtInstance = GetModuleHandleA(crtFileName);
    typedef int(__cdecl *p__getmainargs) (int*, TCHAR***, TCHAR***, int, struct _startupinfo*);
    if (crtInstance == NULL) crtInstance = LoadLibraryA(crtFileName);
    p__getmainargs fn__getmainargs = (crtInstance == NULL) ? NULL : (p__getmainargs)GetProcAddress(crtInstance,
#     ifdef UNICODE
        "__wgetmainargs"
#     else
        "__getmainargs"
#     endif
    );
    if (fn__getmainargs == NULL || fn__getmainargs(&argc, &argv, &envp, 0, &si) != 0) return -1;
  }
  /*** end of workaround ***/


  if (argc != 3 && !(argc == 2 && _tcsicmp(argv[1], TEXT("/l")) == 0))
  {
    Help(argv[0]);
    return 1;
  }

  CALLBACKDATA data = {0};
  uint64_t uptr = 0;
  TCHAR *endptr = NULL;

  switch((*argv[1] != '/' || argv[1][1] == 0 || argv[1][2] != 0) ? 0 : (int)_totupper(argv[1][1]))
  {
    case 'L':
      EnumWindows(EnumWindowsListCallback, (LPARAM)&data);
      return 0;
    case 'H':
      errno = 0;
      uptr = _tcstoui64(argv[2], &endptr, 16);
      if (errno == 0 && *endptr == 0 && (ULONG_PTR)uptr == uptr)
      {
        data.hwnd = (HWND)(ULONG_PTR)uptr;
        data.tid = GetWindowThreadProcessId(data.hwnd, &(data.pid));
        LRESULT length = SendMessage(data.hwnd, WM_GETTEXTLENGTH, 0, 0);
        if (!GetWindow(data.hwnd, GW_OWNER) && PidToPname(&data) && (data.title = (TCHAR*)calloc(length + 1, sizeof(TCHAR))))
        {
          SendMessageTimeout(data.hwnd, WM_GETTEXT, length + 1, (LPARAM)data.title, SMTO_BLOCK | SMTO_ABORTIFHUNG | SMTO_ERRORONEXIT, 2000u, NULL);
          if (SendMessageTimeout(data.hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000u, NULL) != 0)
          {
            PrintData(&data);
            free(data.title);
            return 0;
          }
        }
      }
      break;
    case 'N':
      _tcscpy(data.name, argv[2]);
      if (PnameToPid(&data))
      {
        EnumWindows(EnumWindowsPIDCallback, (LPARAM)&data);
        if (data.hwnd && SendMessageTimeout(data.hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000u, NULL) != 0)
        {
          PrintData(&data);
          if (data.title)
            free(data.title);
          return 0;
        }
      }
      break;
    case 'P':
      errno = 0;
      data.pid = _tcstoul(argv[2], &endptr, 0);
      if (errno == 0 && *endptr == 0)
      {
        EnumWindows(EnumWindowsPIDCallback, (LPARAM)&data);
        if (data.hwnd && PidToPname(&data) && SendMessageTimeout(data.hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000u, NULL) != 0)
        {
          PrintData(&data);
          if (data.title)
            free(data.title);
          return 0;
        }
      }
      break;
    case 'T':
      errno = 0;
      data.tid = _tcstoul(argv[2], &endptr, 0);
      if (errno == 0 && *endptr == 0)
      {
        EnumWindows(EnumWindowsTIDCallback, (LPARAM)&data);
        if (data.hwnd && PidToPname(&data) && SendMessageTimeout(data.hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000u, NULL) != 0)
        {
          PrintData(&data);
          if (data.title)
            free(data.title);
          return 0;
        }
      }
      break;
    case 'W':
      if (*argv[2])
      {
        data.title = argv[2];
        EnumWindows(EnumWindowsTitleCallback, (LPARAM)&data);
        if (data.hwnd && PidToPname(&data) && SendMessageTimeout(data.hwnd, WM_CLOSE, 0, 0, SMTO_BLOCK | SMTO_NOTIMEOUTIFNOTHUNG, 2000u, NULL) != 0)
        {
          PrintData(&data);
          return 0;
        }
      }
      break;
    default:
      Help(argv[0]);
  }

  return 1;
}

BOOL PnameToPid(CALLBACKDATA *const pData)
{
  HANDLE hProcessSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  PROCESSENTRY32 procentry = {0};
  procentry.dwSize = sizeof(PROCESSENTRY32);
  pData->pid = 0;

  if (Process32First(hProcessSnap, &procentry))
  {
    do
    {
      if (_tcsicmp(procentry.szExeFile, pData->name) == 0)
      {
        pData->pid = procentry.th32ProcessID;
        _tcscpy(pData->name, procentry.szExeFile);
      }
    } while (pData->pid == 0 && Process32Next(hProcessSnap, &procentry));
  }

  if (hProcessSnap != INVALID_HANDLE_VALUE)
    CloseHandle(hProcessSnap);

  if (pData->pid == 0)
    return FALSE;

  return TRUE;
}

BOOL PidToPname(CALLBACKDATA *const pData)
{
  HANDLE hProcessSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  PROCESSENTRY32 procentry = {0};
  procentry.dwSize = sizeof(PROCESSENTRY32);
  *(pData->name) = 0;

  if (Process32First(hProcessSnap, &procentry))
  {
    do
    {
      if (pData->pid == procentry.th32ProcessID)
        _tcscpy(pData->name, procentry.szExeFile);
    } while (*(pData->name) == 0 && Process32Next(hProcessSnap, &procentry));
  }

  if (hProcessSnap != INVALID_HANDLE_VALUE)
    CloseHandle(hProcessSnap);

  if (*(pData->name) == 0)
    return FALSE;

  return TRUE;
}

BOOL CALLBACK EnumWindowsListCallback(HWND hwnd, LPARAM lParam)
{
  CALLBACKDATA *pData = (CALLBACKDATA*)lParam;
  pData->hwnd = hwnd;
  pData->tid = GetWindowThreadProcessId(hwnd, &(pData->pid));
  LRESULT length = SendMessage(hwnd, WM_GETTEXTLENGTH, 0, 0);
  if (!GetWindow(hwnd, GW_OWNER) && PidToPname(pData) && (pData->title = (TCHAR*)calloc(length + 1, sizeof(TCHAR))))
  {
    SendMessageTimeout(hwnd, WM_GETTEXT, length + 1, (LPARAM)pData->title, SMTO_BLOCK | SMTO_ABORTIFHUNG | SMTO_ERRORONEXIT, 2000u, NULL);
    PrintData(pData);
    free(pData->title);
  }

  return TRUE;
}

BOOL CALLBACK EnumWindowsPIDCallback(HWND hwnd, LPARAM lParam)
{
  CALLBACKDATA *pData = (CALLBACKDATA*)lParam;
  static DWORD process_id = 0;
  pData->tid = GetWindowThreadProcessId(hwnd, &process_id);
  if (pData->pid == process_id && !GetWindow(hwnd, GW_OWNER))
  {
    pData->hwnd = hwnd;
    pData->title = NULL;
    LRESULT length = SendMessage(hwnd, WM_GETTEXTLENGTH, 0, 0);
    if ((pData->title = (TCHAR*)calloc(length + 1, sizeof(TCHAR))))
      SendMessageTimeout(hwnd, WM_GETTEXT, length + 1, (LPARAM)pData->title, SMTO_BLOCK | SMTO_ABORTIFHUNG | SMTO_ERRORONEXIT, 2000u, NULL);

    return FALSE;
  }

  return TRUE;
}

BOOL CALLBACK EnumWindowsTIDCallback(HWND hwnd, LPARAM lParam)
{
  CALLBACKDATA *pData = (CALLBACKDATA*)lParam;
  static DWORD thread_id = 0;
  thread_id = GetWindowThreadProcessId(hwnd, &(pData->pid));
  if (pData->tid == thread_id && !GetWindow(hwnd, GW_OWNER))
  {
    pData->hwnd = hwnd;
    pData->title = NULL;
    LRESULT length = SendMessage(hwnd, WM_GETTEXTLENGTH, 0, 0);
    if ((pData->title = (TCHAR*)calloc(length + 1, sizeof(TCHAR))))
      SendMessageTimeout(hwnd, WM_GETTEXT, length + 1, (LPARAM)pData->title, SMTO_BLOCK | SMTO_ABORTIFHUNG | SMTO_ERRORONEXIT, 2000u, NULL);

    return FALSE;
  }

  return TRUE;
}

BOOL CALLBACK EnumWindowsTitleCallback(HWND hwnd, LPARAM lParam)
{
  CALLBACKDATA *pData = (CALLBACKDATA*)lParam;
  static TCHAR *tmp = NULL;

  LRESULT length = SendMessage(hwnd, WM_GETTEXTLENGTH, 0, 0);

  if (length && (tmp = (TCHAR*)malloc((length + 1) * sizeof(TCHAR))))
  {
    if (SendMessageTimeout(hwnd, WM_GETTEXT, length + 1, (LPARAM)tmp, SMTO_BLOCK | SMTO_ABORTIFHUNG | SMTO_ERRORONEXIT, 2000u, NULL) != 0
        && _tcscmp(tmp, pData->title) == 0
        && !GetWindow(hwnd, GW_OWNER))
    {
      pData->hwnd = hwnd;
      pData->tid = GetWindowThreadProcessId(hwnd, &(pData->pid));
      free(tmp);
      return FALSE;
    }
    free(tmp);
  }

  return TRUE;
}

void PrintData(const CALLBACKDATA *const pData)
{
  _tprintf(TEXT("%p|%lu|%lu|%s|%s\n"), pData->hwnd, pData->pid, pData->tid, pData->name, pData->title ? pData->title : TEXT(""));
}

void Help(TCHAR *const owncall)
{
  TCHAR *name = _tcsrchr(owncall, TEXT('\\'));
  name = name ? name + 1 : owncall;
  TCHAR *name2 = _tcsrchr(name, TEXT('/'));
  if (name2) name = name2 + 1;
  TCHAR *p = _tcsrchr(name, TEXT('.'));
  if (p && _tcsicmp(p, TEXT(".exe")) == 0) *p = 0;

  _tprintf(TEXT(
    "Tries to close the main window of a process by sending the WM_CLOSE message.\n\n"
    "Usage:\n"
    "%s /L | /H HWND | /N \"process name\" | /P PID | /T TID | /W \"window title\"\n\n"
    "  option /L      Creates a list of all main windows.\n"
    "                 (window handle|process ID|process name|window title)\n"
    "  HWND           Hexadecimal window handle as displayed using option /L.\n"
    "  process name   Name (including .exe extension) of the process owning the window.\n"
    "  PID            Process ID of the process owning the window.\n"
    "  TID            Thread ID of the thread owning the window.\n"
    "  window title   Window title as displayed in the task bar.\n\n"
    "Lists window handle, process ID, process name and window title of the window whereto\n"
    "WM_CLOSE was sent.\n\n"
    "Returns 0 if the WM_CLOSE message was successfully sent to the found window. Otherwise 1\n"
    "will be returned.\n\n"
    "NOTE: Sending WM_CLOSE to a window does neither guarantee that the window will be closed\n"
    "nor that the process will be terminated.\n\n"
  ), name);
}
